
package br.remoto.model.Neuron;

import java.util.ArrayList;

import br.remoto.model.Nerve;
import br.remoto.model.ReMoto;
import br.remoto.model.vo.NeuronVO;


public class Motoneuron extends SpinalNeuron
{
	private static final long serialVersionUID = 1L;

	protected double tAxonSpike;

	protected double axonThreshold;

	protected double latencyStimulusSpinal;
	protected double latencyStimulusEndPlate;
	protected double latencySpinalEndPlate;

	protected ArrayList orthodromicSpikeTrain;
	protected ArrayList antidromicSpikeTrain;

	protected int indexAntidromicSpike;
	protected int indexOrthodromicSpike;
	
	protected Nerve nerve;

	
	public Motoneuron()
	{
	}
	
	
	public Motoneuron(NeuronVO neu, int index, int indexCategory, Miscellaneous misc)
	{
		super(neu, index, misc);
		this.index = indexCategory;

		reset();
	}

	
	public void reset()
	{
		super.reset();
		
		tAxonSpike = -ReMoto.T_MAX;

		if( orthodromicSpikeTrain != null ) 
			orthodromicSpikeTrain.clear();
		else
			orthodromicSpikeTrain = new ArrayList();
		
		if( antidromicSpikeTrain != null ) 
			antidromicSpikeTrain.clear();
		else
			antidromicSpikeTrain = new ArrayList();
		
		indexAntidromicSpike = 0;
		indexOrthodromicSpike = 0;
	}

	
	public void atualize(double t) 
	{
		try
		{
			super.atualize(t);
	
			atualizeDendrite(t);
			atualizeSoma(t);
			atualizeAxon(t);
			
			double tAntidromicSpike = ReMoto.T_MAX;
			double tOrthodromicSpike = -ReMoto.T_MAX;
			
			// Verify antidromic propagation of axonal spikes and collision
			if( indexAntidromicSpike < antidromicSpikeTrain.size() )
			{
				tAntidromicSpike = ((Double)antidromicSpikeTrain.get(indexAntidromicSpike)).doubleValue();
				
				if( t > tAntidromicSpike + latencyStimulusSpinal )
				{
					// Verify collision between soma spike and axon spike
					if( t - tSomaSpike >= miscellaneous.getMnSomaRefractoryPeriod() - ReMoto.T_TOLERANCE )
					{
						addSomaSpike(t);
					}
	
					indexAntidromicSpike++;
				}
			}
	
			// Verify orthodromic propagation of soma spikes and collision
			if( indexOrthodromicSpike < orthodromicSpikeTrain.size() )
			{
				tOrthodromicSpike = ((Double)orthodromicSpikeTrain.get(indexOrthodromicSpike)).doubleValue();
					
				if( t > tOrthodromicSpike + latencyStimulusSpinal )
				{
					// Verify collision between soma spike and axon spike
					if( t - tAxonSpike >= miscellaneous.getMnAxonRefractoryPeriod() - ReMoto.T_TOLERANCE )
					{
						addAxonSpike(t);
					}
	
					indexOrthodromicSpike++;
				}
			}
			
			// Verify collision in the midlle of the axon
			if( Math.abs( tAntidromicSpike - tOrthodromicSpike ) < latencyStimulusSpinal )
			{
				// Increase indexes in order to colide both spikes
				indexAntidromicSpike++;
				indexOrthodromicSpike++;
			}
				
		}
		catch (Exception e) 
		{
			System.out.println( "Error while atualizing MN: " + cd );
		}
	}

	
	protected void atualizeDendrite(double t)
	{
		try
		{
			double k1 = dVd_dt( 1, t,									saturate( vd ) );
			double k2 = dVd_dt( 2, t + miscellaneous.getStepByTwo(),	saturate( vd + miscellaneous.getStepByTwo() * k1 ) );
			double k3 = dVd_dt( 3, t + miscellaneous.getStepByTwo(),	saturate( vd + miscellaneous.getStepByTwo() * k2 ) );
			double k4 = dVd_dt( 4, t + miscellaneous.getStep(),			saturate( vd + miscellaneous.getStep() * k3 ) );

			vd = saturate( vd + miscellaneous.getStepBySix() * (k1 + 2*k2 + 2*k3 + k4) );
		}
		catch(Exception e)
		{
			System.out.println( "MN.atualizeDendrite: " + e.getMessage() );
		}
	}
	
	
	protected void atualizeSoma(double t)
	{
		try
		{
			double k1 = dVs_dt( 1, t,									saturate( vs ) );
			double k2 = dVs_dt( 2, t + miscellaneous.getStepByTwo(),	saturate( vs + miscellaneous.getStepByTwo() * k1 ) );
			double k3 = dVs_dt( 3, t + miscellaneous.getStepByTwo(),	saturate( vs + miscellaneous.getStepByTwo() * k2 ) );
			double k4 = dVs_dt( 4, t + miscellaneous.getStep(),			saturate( vs + miscellaneous.getStep() * k3 ) );
	
			vs = saturate( vs + miscellaneous.getStepBySix() * (k1 + 2*k2 + 2*k3 + k4) );
			
			// Verify soma spike generation
			// Verify colision between soma and axon spikes
			if( vs > threshold  && 
				t - tSomaSpike > miscellaneous.getMnSomaRefractoryPeriod() )
			{
				addSomaSpike(t);
				orthodromicSpikeTrain.add( new Double(t) );
			}
		}
		catch(Exception e)
		{
			System.out.println( "MN.atualizeSoma: " + e.getMessage() );
		}
	}

	
	public void atualizeAxon(double t)
	{
		if( nerve == null || nerve.isActive() == false )
			return;
		
		// ??? teste de inibicao reciproca homonima
		//if( cd.equals( "MN S 125" ) || cd.equals( "MN FR 62" ) || cd.equals( "MN FF 62" ) )
		//	return;

		// Verify axon spike generation
		// Verify colision between axon and soma spikes
		if( nerve.getIntensity() > axonThreshold && 
			t - tAxonSpike > miscellaneous.getMnAxonRefractoryPeriod() )
		{
			addAxonSpike(t);
			antidromicSpikeTrain.add( new Double(t) );
		}
	}


	protected double dVd_dt(int slope, double t, double Vd)
	{
		double iSynaptic = 0;

		// Synapses include all inputs and noises
		iSynaptic += dendExcitSynapses.getCurrent(slope, t, Vd);
		iSynaptic += dendInhibSynapses.getCurrent(slope, t, Vd);
		
		double iSoma = gCoupling * (vs - Vd);
		double iLeak = gLeakDend * (ReMoto.E0 - Vd);

		double iInjected = 0;

		if( current != null && ReMoto.dendriteCompartment.equals( current.getCompartment() ) )
		{
			iInjected = current.getCurrent(t);
		}

		// Currents in nA - capacitance in nF
		double retorno = (iSynaptic + iSoma + iLeak + iInjected) / capacitanceDend;

		return retorno;
	}


	protected double dVs_dt(int slope, double t, double Vs)
	{
		double iIonic = 0;
		double iSynaptic = 0;

		iIonic += gNa.getCurrent(slope, t, Vs);
		iIonic += gKf.getCurrent(slope, t, Vs);
		iIonic += gKs.getCurrent(slope, t, Vs);

		// Synapses include all inputs and noises
		iSynaptic += somaExcitSynapses.getCurrent(slope, t, Vs);
		iSynaptic += somaInhibSynapses.getCurrent(slope, t, Vs);
		
		double iDendrite = gCoupling * (vd - Vs);
		double iLeak = gLeakSoma * (ReMoto.E0 - Vs);
		
		double iInjected = 0;

		if( current != null && ReMoto.somaCompartment.equals( current.getCompartment() ) )
		{
			iInjected = current.getCurrent(t);
		}

		// Currents in nA - capacitance in nF
		double retorno = (iIonic + iSynaptic + iDendrite + iLeak + iInjected) / capacitanceSoma;

		return retorno;
	}

	
	public void addSomaSpike(double t)
	{
		tSomaSpike = t;
		somaSpikeTrain.add(new Double(t));
		
		gNa.start(t);
		gKf.start(t);
		gKs.start(t);
		
		propagateSpike(t);
	}

	
	public void addAxonSpike(double t)
	{
		tAxonSpike = t;
		
		terminalSpikeTrain.add(new Double(t + latencyStimulusEndPlate));
	}


	public void setLatencies(double  axonVelocity, double stimulusSpinalDistance, double stimulusEndPlateDistance) 
	{
		if( axonVelocity < 0.01 )
			axonVelocity = 0.01;
		
		latencyStimulusSpinal = (stimulusSpinalDistance/axonVelocity) * 1000.0; 
		latencyStimulusEndPlate = (stimulusEndPlateDistance/axonVelocity) * 1000.0;
		latencySpinalEndPlate = ((stimulusSpinalDistance + stimulusEndPlateDistance)/axonVelocity) * 1000.0;
	}


	public double getLatencySpinalEndPlate() {
		return latencySpinalEndPlate;
	}


	public void setLatencySpinalEndPlate(double latencySpinalEndPlate) {
		this.latencySpinalEndPlate = latencySpinalEndPlate;
	}


	public double getLatencyStimulusEndPlate() {
		return latencyStimulusEndPlate;
	}


	public void setLatencyStimulusEndPlate(double latencyStimulusEndPlate) {
		this.latencyStimulusEndPlate = latencyStimulusEndPlate;
	}


	public double getLatencyStimulusSpinal() {
		return latencyStimulusSpinal;
	}


	public void setLatencyStimulusSpinal(double latencyStimulusSpinal) {
		this.latencyStimulusSpinal = latencyStimulusSpinal;
	}


	public double getAxonThreshold() {
		return axonThreshold;
	}


	public void setAxonThreshold(double axonThreshold) {
		this.axonThreshold = axonThreshold;
	}


	public Nerve getNerve() {
		return nerve;
	}


	public void setNerve(Nerve nerve) {
		this.nerve = nerve;
	}


}
